"""
	Steve L. Nyemba <steve@the-phi.com>
	The Phi Technology, LLC - Store
	
	This file handles customer & plans associated with a given product/app
	We understand that a product will have multiple plans under it and we make sure that we always have a free version:
		- Having a free product insures there is no excuse not to signup users
		- If a product doesn't fall under this model we will find a way to fix it.
	- 
	The store must be prefixed with an authentication module, because customers must be signed in before making a purchase.
"""
from __future__ import division
from flask import Flask, request, session, render_template
from flask_session import Session
from flask_cors import CORS
from flask_pymongo import PyMongo
from flask_mongo_sessions import MongoDBSessionInterface

import stripe
import json
# from StringIO import StringIO
import re
import os
from utils.params import PARAMS
from transport import factory
# from store import Store
import store
from datetime import datetime
import numpy as np
import requests

# from utils.transport import Couchdb, CouchdbReader
PORT = 8100 if 'port' not in PARAMS else int(PARAMS['port']) ;
path = PARAMS['path'] #os.environ['CONFIG']
f = open(path)
CONFIG = json.loads(f.read())
CONTEXT= PARAMS['context'] if 'context' in PARAMS else ''
stripe_keys = {
    'secret_key': CONFIG['stripe']['secret'].strip(),
    'publishable_key': CONFIG['stripe']['pub'].strip()
}

stripe.api_key = stripe_keys['secret_key'].strip()


	
app = Flask(__name__)
CORS(app)

SYS_STORE = CONFIG['store']
@app.route("/")
def index():
	mystore = store.factory.instance(name='music')
	products = mystore.get.products()
	args = {"context":CONTEXT,"products":products}
	return json.dumps(products),"content-type:application/json"
	# return render_template("index.html",**args)
@app.route("/json/<id>")
def get_plans_json(id):
	"""
	This function returns the plans for a given product
	"""
	HEADER="Content-type:application/json"
	try:
		mystore = store.factory.instance(name=id)
		return json.dumps(mystore.get.plans()),HEADER
	except Exception as e:
		#
		# Log this shit or not
		pass
	return "[]",HEADER

@app.route("/store",methods=['POST','PUT'])
def to_session():
	"""
	This function will store/persist session variables for reuse
	"""
	info = request.json
	if info :
		for key in info :
			session[key] = info[key]
	return "{}"
@app.route("/signup",methods=['GET','POST'])
def signup():
	"""
	This function will signup a user to a plan or upgrade them if necessary
	The assumption here is one product - one plan
	:request.form	stripeToken (optional)
	"""
	if request.method == 'GET' :
		if 'auth' not in session :
			return "0",403
		_object = {"data-key":CONFIG['stripe']['pub'].strip()}	
		return json.dumps(_object),{"Content-Type":"application/json"}
	else:
		#
		# We are receiving a call from stripe framework and we need create a charge for this
		# @NOTE: https://stripe.com/docs/payments/accept-a-payment-charges#web-create-charge
		token 	= request.form['stripeToken'] if 'stripeToken' in request.form else None
		email 	= session['auth']['user']['uid']
		plan_id	= session['plan']['id']
		id	= session['product']['name']
		mystore = store.factory.instance(name=id,email=email)
		user	= mystore.user
		plans 	= user.plan.info()
		plans	= [plans] if isinstance(plans,dict) else plans
		
		has_plan = [1 for item in plans if item['id'] == plan_id]
		if np.sum(has_plan) == 0 and len(plans) == 0:
			
			mystore.subscribe(id=plan_id,email=email)
		elif np.sum(has_plan)==0 and len(plans) > 0:
			#
			# We should upgrade/downgrade the current plan
			#
			mystore.plan.upgrade(plan_id,email)
		uri = session['product']['metadata']['uri'] if 'uri' in session['product']['metadata'] else '#status'
		
		#
		# @NOTE:
		#	This function assumes one product one plan, operations outside of this scope will require a new function
		return """
			<script>
				window.location = ':uri'
			</script>
		""".replace(":uri",uri)

@app.route("/ui/signup/<id>",methods=['GET','PUT'])
def signup_form(id):
	"""
	This function loads the signup form with cloud-view authentication support
	This allows us to know who is signing up or upgrading their plan
	"""
	
	mystore = store.factory.instance(name=id)
	plans = mystore.get.plans()
	index = int(request.args['index'])
	plan = plans[index]
	
	args = {"product":id,"label":mystore.product['statement_descriptor'],"plan":plan,"context":CONTEXT,"now":datetime.now().year}
	args['theme'] = 'theme-clouds'
	args['cloud-view'] = CONFIG['cloud-view']
	
	return render_template("signup.html",**args)
@app.route("/ui/<id>")
def get_plans_ui(id):
	"""
	This function loads the form for signup and information about the various plans for a given product.
	:id	name of the products
	"""	
	mystore = store.factory.instance(name=id)
	#
	# sort plans by 
	
	description = mystore.product['metadata']['about'] if 'about' in mystore.product['metadata'] else ''
	label = mystore.product['statement_descriptor']
	args = {"product":id,"label":label,"description":description,"context":CONTEXT,"plans":mystore.get.plans(),"now":datetime.now().year}
	args['theme'] = 'theme-clouds'
	# print (mystore.product.keys())
	session['cloud-view'] = CONFIG['cloud-view']
	session['product'] = mystore.product
	return render_template('plans.html',**args)
@app.route("/goto/<id>",methods=['POST','GET'])
def redirect(id ) :
	"""
	This function will redirect to a given product page given a payload. The payload is intended to set the a session
	:info	
	"""
	
	mystore = store.factory.instance(name=id)
	args 	= {"context":CONTEXT,"product":mystore.product}
	HEADER 	= {"Location":mystore.product['metadata']['uri']}
	info	= request.json
	if info :
		session['auth'] = info['auth']
		session['plan'] = info['plan']
		session['product'] = mystore.product
		return "1",HEADER
	else:

		return "0",403
	pass
@app.route("/init/<product>",methods=['POST'])
def init(product):
	"""
		This function initializes/logs a product to a given user, i.e it will create a session for users/product/plans 
		if the user has provided a user identifier it will be used as her primary email. The understanding is that a product may have multiple plans under it but always a free one
		@param uid	user's email (primary)
		@param pid	plan identifier
	"""
	#
	# get product and plans
	#
	uid	= request.headers['uid']
	plan_id = request.headers['pid'] if 'pid' in request.headers else None

	mystore =  store.factory.instance(name=product,email=uid)
	
	
	user = mystore.user
	user.init (uid)
	if not user.plan.info() :
		mystore.plan.subscribe(email = user.info()['email'])
	
	
	features = user.plan.info()['metadata']['features'] if 'features' in user.plan.info()['metadata'] else {}	
	#
	#@TODO: Log that uid, has initiated use of product
	#
	return json.dumps(features),200
	

# @app.route('/subscribe/<product>',methods=['POST'])
# def subscribe(product):
# 	"""
# 		This function subscribes a user to a given service for an application
# 		This function guarantees not to duplicate subscriptions
# 		@resource 	name	name of the application {cloud-music}
# 		@header		key	service/plan
# 	"""

# 	#
# 	# The name is the full name of the service
# 	#
# 	resp = "0"
# 	user = User(stripe=stripe,store=CONFIG['couchdb'],product=product)
# 	if 'stripeToken' in request.form :
# 		stripeToken	= request.form['stripeToken']
# 		uid		= request.form['stripeEmail']
# 		tokenType	= request.form['stripeTokenType']
# 		amount		= request.form['amount']
# 		pid 		= request.form['plan']

# 	else:
# 		pid = request.headers['pid'] if 'pid' in request.headers else None
# 		uid = request.headers['uid']
# 		stripeToken = None
# 	user.subscribe(uid,pid,stripeToken)
# 	if 'auid' in request.headers :
		
# 		if request.headers['auid'].startswith('[') or request.headers['auid'].startswith("{") :
# 			auid = json.loads(request.headers['auid'])
# 		else:
# 			auid = [request.headers['auid']]
# 		user.update(emails=auid)
# 	user.refresh(uid)
	
# 	reader = CouchdbReader(uri=SYS_STORE['uri'],dbname=product,uid=uid,create=False)
# 	plans = reader.view('users/active_plan',key=user.user_key) #me['_id'])
# 	#session['plans'] = plans
# 	session['key'] = user.user_key	
# 	session['uid'] = uid
# 	session['active-plans']  = [item['id'] for item in plans]
# 	print (session['active-plans'])
# 	return (json.dumps(plans),200)
	
def get_plans(product) :
	lproducts = stripe.Product.list()
	lproducts = [item for item in lproducts.auto_paging_iter() if item.name == product ]
	if lproducts :
		product_id = lproducts[0].id
		alias = lproducts[0].statement_descriptor
		plans = stripe.Plan.list(product=product_id)
		i = 0
		for plan in plans :
			plan['product_alias'] = alias if alias is not None else ''
			if 'features' in plan['metadata']:
				plan['metadata']['features'] = json.loads(plan['metadata']['features'])
				plans[i] = plan
			i += 1
		if plans :
			plans = plans.data
		plans.sort(key = lambda x:x.amount)	
		
		return plans
	else:
		return []

@app.route('/features/<product>')
def features(product):
	"""
		This function returns the plan/features of a user for a given application if provided a uid (email)
		if no uid are provided then the function will return all the plans/features associated with the given product
		@header uid	user's email
	"""

	plans = []
	if 'uid' in request.headers :
		uid = request.headers['uid']
		couchdb = CouchdbReader(uri=CONFIG['couchdb']['uri'],dbname=product,uid=uid,create=False)
		key  = couchdb.view("users/uid_map",key=uid)
		if key :
			key = key[0]['value']
			plans = couchdb.view('users/active_plan',key=key)
			plans =  [plan['value'] for plan in plans if 'value' in plan]
		else:
			plans = []
	else:
		plans = get_plans(product)
	#
	#  formatting plans for the output
	# 
	return json.dumps(plans)
"""
	This function returns a user's plans/status for an application
	@header uid 	user's email address
"""
@app.route('/status/<app_name>')
def status(app_name):
	
	uid = request.headers['uid']
	
	couchdb = CouchdbReader(uri=CONFIG['couchdb']['uri'],dbname=app_name,uid=uid,create=False)
	handler = Domain.User(couchdb.dbase,stripe)
	handler.initialize(uid)
	#lsub = handler.subscriptions()
	plans = handler.plans()
	return json.dumps(plans)
"""
	This endpoint is the signup form for a given application,
	It will allow user's to be able to signup for various plans
	@pre 'uid' in request.headers
"""	
@app.route('/signup/<product>')
def _signup(product) :
	apikey = CONFIG['stripe']['pub'].strip()	

	#
	# This function returns the plans for a given application
	# We assume the application name is the prefix of the plan identifier in stripe
	#
	if 'user-info' in session :
		_info = session['user-info']
		session['uid'] = _info['uid']
	uid = session['uid'] if 'uid' in session else ''
	can_purchase = True
	if uid is None :
		uid = request.args.get('uid') if 'uid' in request.args else None
	plans = get_plans(product)
	
	#
	# @TODO: Mark the plans the current user is signed up for
	#
	if 'themes' in CONFIG:
		theme = CONFIG['themes'][product] if product in CONFIG['themes'] else CONFIG['theme']['default']
	else:
		theme = ''
	platform='web' if 'platform' not in request.args else request.args['platform']
	alias = plans[0]['product_alias']
	if 'user-plan' in session :
		user_plans = [item['id'] for item in session['user-plan']]
	else :
		user_plans = []
	# active_plan = session['active-plans']  if 'active-plans' in session else []	
	args = {"context":CONTEXT,"theme":theme,"uid":uid,"alias":alias,"platform":platform,"app_name":product,"apikey":apikey,"plans":plans}
	args['active_plans'] = user_plans

	return render_template('subscribe.html',**args) #context=CONTEXT,uid=uid,alias=alias,platform=platform,app_name=product,plans=plans,apikey=apikey)
	
@app.route('/subscribe/<name>',methods=['DELETE'])
def cancel_subscribe(name) :
	pass
"""
	This function defines if a given user is a customer or not
	We should be able to tell by how we create customers
"""
@app.route('/checkout/<app_name>',methods=['GET'])
def is_customer (app_name):
	uid = request.args.get('uid')
	pid = request.args.get('pid')
	couchdb = CouchdbReader(uri=CONFIG['couchdb']['uri'],dbname=app_name,uid=uid,create=False)
	r = couchdb.view('federation/uid_map',key=uid)
	id = r[0]['value']
	couchdb = CouchdbReader(uri=CONFIG['couchdb']['uri'],dbname=app_name,uid=id,create=False)
	info = couchdb.read()
	lsub = info['subscriptions']
	
	plans = [dict(dict(item['plan'],**{"status":item['status']}),**{"subscription":item['id']}) for item in lsub]
	if pid is not None:
		plans = [item for item in plans if item['id'] == pid]
	#
	# Caching the subscription identifiers so we can create an invoice later on (if need be)
	# @TODO Improve this process later on by allowing user's to pay for what they can (not everything)
	#
	session['plans'] = plans

	session['plans']
	amount = sum([item['amount'] for item in plans])	
	apikey = CONFIG['stripe']['pub'].strip()
	
	amount = amount / 100
	return render_template('bill.html',context=CONTEXT,apikey=apikey,app_name=app_name.replace('-',' '),plans=plans,total_amount=amount)

if __name__ == '__main__' :
	#
	# setup mongodb session management (not sure why)
	app.config['SESSION_TYPE'] = 'mongodb'
	app.config['MONGO_URI'] = 'mongodb://localhost:27017'
	app.config['SESSION_MONGODB_DB'] = 'sessions'
	app.config['SESSION_MONGODB_COLLECT'] = 'store'
	mongo = PyMongo(app)
	app.session_interface = MongoDBSessionInterface(app,mongo.db, 'store')

	app.debug = True ;
	app.secret_key = '360-8y-[0v@t10n]+kr81v17y'
	app.config['MAX_CONTENT_LENGTH'] = 1600 * 1024 * 1024
	Session(app)	
	app.run(port=PORT,threaded=True,host='0.0.0.0')
	# app.run() #'0.0.0.0',PORT,True,threaded=True)
